/* $Id: ifplugd.c 1.12 Sat, 01 Feb 2003 03:00:07 +0100 lennart $ */

/*
 * This file is part of ifplugd.
 *
 * ifplugd is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ifplugd is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ifplugd; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/stat.h>
#include <getopt.h>
#include <stdarg.h>
#include <syslog.h>
#include <fcntl.h>
#include <linux/kd.h>
#include <sys/ioctl.h>
#include <limits.h>
#include <net/if.h>
#include <linux/sockios.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/time.h>

#include "ethtool-local.h"
#include "interface.h"
#include "version.h"

#define VARRUN "/var/run"
#define IFPLUGD_ENV_PREVIOUS "IFPLUGD_PREVIOUS"
#define IFPLUGD_ENV_CURRENT "IFPLUGD_CURRENT"

int interface_auto_up = 1, interface_do_message = 1;

char *interface = "eth0";
char *run = "/etc/ifplugd/ifplugd.action";
char *extra_arg = NULL;
int polltime = 1, delay_up = 0, delay_down = 5;
int daemonize = 1, use_syslog = 1, use_beep = 1, no_shutdown_script = 0, ignore_failure = 0, wait_on_fork = 0;

int wait_on_fork_pipe[2];

enum { API_AUTO, API_ETHTOOL, API_MII, API_PRIVATE } api_mode = API_AUTO;
interface_status_t (*detect_beat_func)(int, char*);

volatile int quit = 0;
volatile int n_sigchld = 0;

void message(int prio, char *format, ...) {
    va_list arglist;
    va_start(arglist, format);
    if (use_syslog) {
	openlog("ifplugd", LOG_PID|LOG_NDELAY, LOG_DAEMON);
        vsyslog(prio, format, arglist);
	closelog();
    } else {
        vfprintf(stderr, format, arglist);
        fprintf(stderr, "\n");
    }
    va_end(arglist);
}

// 0: high, 1: low, 2: very low
void beep(int b) {
    int fd;
    int argp;

    if (!use_beep)
        return;
    
    if ((fd = open("/dev/tty1", O_WRONLY)) < 0) {
        use_beep = 0;
        message(LOG_WARNING, "Could not open /dev/tty, cannot beep.");
        return;
    }

    switch (b) {
        case 0: argp = (100<<16) + 0x637; break;
        case 1: argp = (100<<16) + 0x937; break;
        default: argp = (100<<16) + 0x1237; break;
    }

    if (ioctl(fd, KDMKTONE, argp) != 0) {
        use_beep = 0;
        message(LOG_WARNING, "Beep failure, disabled.");
        return;
    }

    usleep((argp >> 16)*1000);
    
    close(fd);
}

char *pid_file() {
    static char fn[PATH_MAX];
    snprintf(fn, sizeof(fn), "%s/ifplugd.%s.pid", VARRUN, interface);
    return fn;
}

pid_t check_running() {
    char *fn;
    static char txt[256];
    FILE *f;
    pid_t pid;

    if (!(f = fopen(fn = pid_file(), "r")))
        return 0;

    if (!(fgets(txt, sizeof(txt), f))) {
        fprintf(stderr, "PID file corrupt, removing. (%s)\n", fn);
        unlink(fn);
        return 0;
    }
        
    if ((pid = atoi(txt)) == 0) {
        fprintf(stderr, "PID file corrupt, removing. (%s)\n", fn);
        unlink(txt);
        return 0;
    }

    if (kill(pid, 0) != 0 && errno != EPERM) {
        fprintf(stderr, "Daemon %u killed: %s; removing PID file. (%s)\n", pid, strerror(errno), fn);
        unlink(fn);
        return 0;
    }
    
    return pid;
}

int kill_daemon() {
    pid_t pid;
    
    if (!(pid = check_running())) {
        fprintf(stderr, "Daemon for interface %s not running.\n", interface);
        return EXIT_FAILURE;
    }

    if (kill(pid, SIGINT) != 0) {
        fprintf(stderr, "Could not signal daemon %u: %s\n", pid, strerror(errno));
        return EXIT_FAILURE;
    }

    return 0;
}

int create_pid_file() {
    char *fn;
    FILE *f;

    if (!(f = fopen((fn = pid_file()), "w")))
        return -1;

    fprintf(f, "%u", getpid());
    fclose(f);

    return 0;
}

void sigchld(int sig) {
    n_sigchld++;
}

int can_read(int fd) {
    fd_set fds;
    struct timeval tv = {0, 0};

    FD_ZERO(&fds);
    FD_SET(fd, &fds);

    return select(FD_SETSIZE, &fds, 0, 0, &tv) == 1;
}

int action(interface_status_t status) {
    pid_t pid;
    void (*_sigchld)(int);
    int _pipe[2];
    int r = 0;
    unsigned n = 0;
    static char buf[256];
    char *arg = (status == IFSTATUS_UP ? "up" : "down");

    message(LOG_INFO, "Executing '%s %s %s'.", run, interface, arg);
    
    n_sigchld = 0;
    _sigchld = signal(SIGCHLD, sigchld);
    siginterrupt(SIGCHLD, 0);
    
    if (pipe(_pipe) < 0) {
        message(LOG_ERR, "pipe() failed: %s", strerror(errno));
        return -1;
    }

    if ((pid = fork()) < 0) {
        message(LOG_ERR, "fork() failed: %s", strerror(errno));
        return -1;
    } else if (pid == 0) {
        dup2(_pipe[1], 1);
        dup2(_pipe[1], 2);
        
        close(_pipe[0]);
        close(_pipe[1]);
        
        execl(run, run, interface, arg, extra_arg, 0);

        _exit(EXIT_FAILURE);
    }

    close(_pipe[1]);

    n = 0;
    
    while (can_read(_pipe[0]) || (!quit && n_sigchld == 0)) {
        char c;

        if (read(_pipe[0], &c, 1) != 1)
            break;

        buf[n] = c;

        if (c == '\n' || n >= sizeof(buf) - 2) {
	    if (c != '\n') n++;
            buf[n] = 0;

            if (buf[0])
                message(LOG_WARNING, "client: %s", buf);
            
            n = 0;
        } else
            n++;
    }

    if (n > 0) {
        buf[n] = 0;
        message(LOG_WARNING, "client: %s", buf);
    }

    waitpid(pid, &r, 0);

    signal(SIGCHLD, _sigchld); 
    siginterrupt(SIGCHLD, 1);

    close(_pipe[0]);

    

    if (!WIFEXITED(r) || WEXITSTATUS(r) != 0) {
        if (status == IFSTATUS_UP)
            beep(2);
        message(LOG_ERR, "Program execution failed, return value is %i.", WEXITSTATUS(r));
        return -1;
    } else {
        if (status == IFSTATUS_UP)
            beep(0);

        message(LOG_INFO, "Program executed successfully.");
    }
    
    return 0;
}

interface_status_t detect_beat_auto(int fd, char *iface) {
    interface_status_t status;

    if ((status = interface_detect_beat_ethtool(fd, iface)) == IFSTATUS_ERR)
        if ((status = interface_detect_beat_mii(fd, iface)) == IFSTATUS_ERR)
            status = interface_detect_beat_priv(fd, iface);
    
    return status;
}

char *strstatus(interface_status_t s) {
    switch(s) {
        case IFSTATUS_UP: return "up";
        case IFSTATUS_DOWN: return "down";
        case IFSTATUS_ERR: return "error";
        default: return "disabled";
    }
}

interface_status_t detect_beat(int fd, char*iface) {
    interface_status_t status;
    static interface_status_t last_status = (interface_status_t) -1;

    status = detect_beat_func(fd, iface);

    if (status != last_status) {
        setenv(IFPLUGD_ENV_PREVIOUS, strstatus(last_status), 1);
        setenv(IFPLUGD_ENV_CURRENT, strstatus(status), 1);
        last_status = status;
    }
    
    if (status == IFSTATUS_ERR && ignore_failure)
        return IFSTATUS_DOWN;

    return status;
}



int open_iface(char *iface) {
    int fd;
    struct ifreq ifr;
    struct ethtool_drvinfo drvinfo;
    char txt[256];

    if ((fd = interface_open(iface)) < 0) {
        message(LOG_ERR, "Could not create socket: %s", strerror(errno));
        return fd;
    }

    if (interface_auto_up)
        interface_up(fd, iface);
    
    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name)-1);
    
    if (ioctl(fd, SIOCGIFHWADDR, &ifr) == -1) {
        if (ignore_failure) {
            message(LOG_INFO, "Cannot get hardware address, ignoring: %s", strerror(errno));
            return fd;
        }
        
        message(LOG_ERR, "Could not get hardware address: %s", strerror(errno));
        close(fd);
        return -1;
    }

    snprintf(txt, sizeof(txt)-1, "Using interface %s/%02X:%02X:%02X:%02X:%02X:%02X", iface, ifr.ifr_hwaddr.sa_data[0] & 0xFF, ifr.ifr_hwaddr.sa_data[1] & 0xFF, ifr.ifr_hwaddr.sa_data[2] & 0xFF, ifr.ifr_hwaddr.sa_data[3] & 0xFF, ifr.ifr_hwaddr.sa_data[4] & 0xFF, ifr.ifr_hwaddr.sa_data[5] & 0xFF);

    memset(&ifr, 0, sizeof(ifr));
    strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name)-1);

    drvinfo.cmd = ETHTOOL_GDRVINFO;
    ifr.ifr_data = (caddr_t) &drvinfo;
    
    if (ioctl(fd, SIOCETHTOOL, &ifr) != -1)
        message(LOG_INFO, "%s with driver %s %s", txt, drvinfo.driver, drvinfo.version);
    else
        message(LOG_INFO, "%s", txt);
              
    return fd;
}


void sigint(int sig) {
    quit = 1;
}

void work() {
    interface_status_t status;
    int countdown = 0, fd;

    if (create_pid_file() < 0) {
        message(LOG_ERR, "Could not create PID file %s.", pid_file());
        goto finish;
    }
    
    signal(SIGINT, sigint);
    signal(SIGTERM, sigint);
    siginterrupt(SIGINT, 1);
    siginterrupt(SIGTERM, 1);
    siginterrupt(SIGCHLD, 0);

    switch (api_mode) {
        case API_ETHTOOL: detect_beat_func = interface_detect_beat_ethtool; break;
        case API_MII: detect_beat_func = interface_detect_beat_mii; break;
        case API_PRIVATE: detect_beat_func = interface_detect_beat_priv; break;
        default: detect_beat_func = detect_beat_auto; break;
    }

    if ((fd = open_iface(interface)) < 0)
        goto finish;

    if ((status = detect_beat(fd, interface)) == IFSTATUS_ERR)
        goto cleanup;
    
    message(LOG_INFO, "ifplugd "VERSION" successfully initialized, link beat %sdetected.", status == IFSTATUS_UP ? "" : "not ");
    beep(status == IFSTATUS_UP ? 0 : 1);

    interface_do_message = 0;
    
    if (status == IFSTATUS_UP)
        if (action(status))
            goto cleanup;

    if (daemonize && wait_on_fork) {
        char c = status == IFSTATUS_UP ? 2 : (status == IFSTATUS_DOWN ? 3 : 1);
        write(wait_on_fork_pipe[1], &c, 1);
        close(wait_on_fork_pipe[1]);
    }

    while (!quit) {
        interface_status_t r;

        sleep(polltime);

        if (quit)
            break;

        if ((r = detect_beat(fd, interface)) == IFSTATUS_ERR)
            goto cleanup;
        else if (status != r) {
            status = r;

            message(LOG_INFO, "Link beat %s.", status == IFSTATUS_DOWN ? "lost" : "detected");
            beep(status == IFSTATUS_UP ? 0 : 1);
            
            if (countdown)
                countdown = 0;
            else {
                if (status == IFSTATUS_UP)
                    countdown = delay_up/polltime;
                
                if (status == IFSTATUS_DOWN)
                    countdown = delay_down/polltime;
                
                if (!countdown)
                    if (action(status))
                        goto cleanup;
            }
        } else if (countdown) {
            countdown--;

            if (!countdown)
                if (action(status))
                    goto cleanup;
        }
    }

cleanup:
    if (!no_shutdown_script && status == IFSTATUS_UP) {
        setenv(IFPLUGD_ENV_PREVIOUS, strstatus(status), 1);
        setenv(IFPLUGD_ENV_CURRENT, strstatus(-1), 1);
        action(IFSTATUS_DOWN);
        beep(1);
    }
    
    close(fd);

finish:

    unlink(pid_file());
    
    if (use_syslog)
        message(LOG_INFO, "Exit.");
}

void usage(char *p) {
    char *m;

    switch (api_mode) {
        case API_ETHTOOL: m = "ethtool"; break;
        case API_MII: m = "mii"; break;
        case API_PRIVATE: m = "priv"; break;
        default: m = "auto";
    }
    
    if (strrchr(p, '/'))
        p = strchr(p, '/')+1;

    printf("%s [options]\n"
           "   -a --no-auto         Do not enable interface automatically (%s)\n"
           "   -n --no-daemon       Do not daemonize (for debugging) (%s)\n"
           "   -s --no-syslog       Do not use syslog, use stderr instead (for debugging) (%s)\n"
           "   -b --no-beep         Do not beep (%s)\n"
           "   -f --ignore-fail     Ignore detection failure, retry instead (for PCMCIA) (%s)\n"
           "   -i --iface=IFACE     Specify ethernet interface (%s)\n"
           "   -r --run=EXEC        Specify program to execute (%s)\n"
           "   -t --poll-time=SECS  Specify poll time in seconds (%i)\n"
           "   -u --delay-up=SECS   Specify delay for configuring interface (%i)\n"
           "   -d --delay-down=SECS Specify delay for deconfiguring interface (%i)\n"
           "   -m --api-mode=MODE   Force API mode (mii, priv, ethtool, auto) (%s)\n"
           "   -q --no-shutdown     Don't run script on daemon quit (%s)\n"
           "   -w --wait-on-fork    Wait until daemon fork finished (%s)\n"
           "   -h --help            Show this help\n"
           "   -k --kill            Kill a running daemon\n"
	   "   -c --check-running   Check if a daemon is currently running\n"
           "   -v --version         Show version\n",
           p, !interface_auto_up ? "on" : "off", !daemonize ? "on" : "off", !use_syslog ? "on" : "off", !use_beep ? "on" : "off", ignore_failure ? "on" : "off",
           interface, run, polltime, delay_up, delay_down, m, no_shutdown_script ? "on" : "off", wait_on_fork ? "on" : "off");
}

void parse_args(int argc, char *argv[]) {
    static struct option long_options[] = {
        {"no-auto",      no_argument, 0, 'a'},
        {"no-daemon",    no_argument, 0, 'n'},
        {"no-syslog",    no_argument, 0, 's'},
        {"no-beep",      no_argument, 0, 'b'},
        {"ignore-fail",  no_argument, 0, 'f'},
        {"iface",        required_argument, 0, 'i'},
        {"run",          required_argument, 0, 'r'},
        {"poll-time",    required_argument, 0, 't'},
        {"delay-up",     required_argument, 0, 'u'},
        {"delay-down",   required_argument, 0, 'd'},
        {"api-mode",     required_argument, 0, 'm'},
        {"wait-on-fork", no_argument, 0, 'w'},
        {"no-shutdown",  no_argument, 0, 'q'},
        {"help",         no_argument, 0, 'h'},
        {"kill",         no_argument, 0, 'k'},
        {"check-running",no_argument, 0, 'c'},
        {"version",      no_argument, 0, 'v'},
        {"extra-arg",    required_argument, 0, 'x'},
        {0, 0, 0, 0}
    };
    int option_index = 0;
    int help = 0, _kill = 0, _check = 0, _version = 0;
    
    for (;;) {
        int c;
        
        if ((c = getopt_long(argc, argv, "asni:r:t:u:d:hkbfvm:qwx:c", long_options, &option_index)) < 0)
            break;

        switch (c) {
            case 'a' :
                interface_auto_up = !interface_auto_up;
                break;
            case 's' :
                use_syslog = !use_syslog;
                break;
            case 'n' :
                daemonize = !daemonize;
                break;
            case 'i' :
                interface = strdup(optarg);
                break;
            case 'r':
                run = strdup(optarg);
                break;
            case 't':
                polltime = atoi(optarg);
                if (polltime < 0) polltime = 0;
                break;
            case 'u':
                delay_up = atoi(optarg);
                break;
            case 'd':
                delay_down = atoi(optarg);
                break;
            case 'h':
                help = 1;
                break;
            case 'k':
                _kill = 1;
                break;
            case 'c':
                _check = 1;
                break;
            case 'v':
                _version = 1;
                break;
            case 'b':
                use_beep = !use_beep;
                break;
            case 'f':
                ignore_failure = !ignore_failure;
                break;
            case 'm':
                switch (tolower(optarg[0])) {
                    case 'e': api_mode = API_ETHTOOL; break;
                    case 'm': api_mode = API_MII; break;
                    case 'p': api_mode = API_PRIVATE; break;
                    case 'a': api_mode = API_AUTO; break;
                    default:
                        fprintf(stderr, "Unknown API mode: %s\n", optarg);
                        exit(2);
                }
                break;
            case 'q':
                no_shutdown_script = !no_shutdown_script;
                break;
            case 'w':
                wait_on_fork = !wait_on_fork;
                break;
            case 'x':
                extra_arg = strdup(optarg);
                break;
            default:
                fprintf(stderr, "Unknown parameter.\n");
                exit(1);
        }
    }

    if (help) {
        usage(argv[0]);
        exit(0);
    }

    if (_kill)
        exit(kill_daemon());

    if (_version) {
        printf("ifplugd "VERSION" (PRCS $ProjectHeader: ifplugd 0.40 Sat, 01 Feb 2003 03:03:31 +0100 lennart $)\n");
        exit(0);
    }

    if (_check) {
        pid_t pid = check_running();

        if (pid == 0)
            printf("ifplugd not running.\n");
        else
            printf("ifplugd process for device %s running as pid %u.\n", interface, pid);
        exit(pid == 0 ? 255 : 0);
    }
    
}

static volatile int alarmed = 0;

void sigalrm() {
    alarmed = 1;
}

int main(int argc, char* argv[]) {

    parse_args(argc, argv);

    if (geteuid() != 0) {
        fprintf(stderr, "Sorry, you need to be root to run this binary.\n");
        return 2;
    }

    if (check_running()) {
        fprintf(stderr, "Sorry, there is already an instance of ifplugd for %s running.\n", interface);
        return 4;
    }
    
    if (daemonize) {
        pid_t pid;

        if (wait_on_fork) {
            if (pipe(wait_on_fork_pipe)) {
                fprintf(stderr, "Sorry, could not create pipe: %s\n", strerror(errno));
                return 5;
            }
        }
        
        if ((pid = fork()) < 0) { // First fork
            fprintf(stderr, "fork() failed: %s\n", strerror(errno));
            return 3;
        } else if (pid == 0) { // First child
            
            if ((pid = fork()) < 0) { // Second fork
                message(LOG_ERR, "fork() failed: %s", strerror(errno));
                return 3;
            } else if (pid > 0) {
                close(wait_on_fork_pipe[0]);
                close(wait_on_fork_pipe[1]);
                return 0; // Second father
            }

            // Detach from everything 
            close(0); 
            close(1); 
            close(2); 

            if (open("/dev/null", O_RDONLY) != 0)
                message(LOG_ERR, "open(\"/dev/null\", O_RDONLY)=0 failed: %s", strerror(errno));
	    if (open("/dev/null", O_WRONLY) != 1)
                message(LOG_ERR, "open(\"/dev/null\", O_WRONLY)=1 failed: %s", strerror(errno));
	    if (open("/dev/null", O_WRONLY) != 2)
                message(LOG_ERR, "open(\"/dev/null\", O_WRONLY)=2 failed: %s", strerror(errno));
            
            setsid();
            umask(777);
            chdir("/");

            // Close the other side of the pipe
            if (wait_on_fork)
                close(wait_on_fork_pipe[0]);
        } else { // First father
            char c = 1;
            if (wait_on_fork) {
                ssize_t r;
                
                signal(SIGALRM, sigalrm);
                siginterrupt(SIGALRM, 1);
                alarm(60);

                if ((r = read(wait_on_fork_pipe[0], &c, 1)) != 1) {

                    if (r < 0 && alarmed)
                        fprintf(stderr, "Timeout reached, killing background process.\n");
                    else
                        fprintf(stderr, "Communication with background process failed, killing: %s\n", r < 0 ? strerror(errno) : "eof");

                    kill(pid, SIGTERM);
                }
                
                signal(SIGALRM, SIG_DFL);

                close(wait_on_fork_pipe[0]);
                close(wait_on_fork_pipe[1]);
            }
            return c;
        }
    }


    // Let's go
    work();
    return 0;
}
